#include <msp/datafile/collection.h>
#include <msp/strings/format.h>
#include "error.h"
#include "material.h"
#include "renderpass.h"
#include "program.h"
#include "programdata.h"
#include "renderer.h"
#include "texture.h"
#include "texture2d.h"
#include "texturing.h"

using namespace std;

namespace Msp {
namespace GL {

RenderPass::RenderPass():
	shprog(0),
	shdata(0),
	material(0),
	texturing(0),
	back_faces(false)
{ }

RenderPass::RenderPass(const RenderPass &other):
	shprog(other.shprog),
	shdata(other.shdata ? new ProgramData(*other.shdata) : 0),
	material(other.material),
	texturing(other.texturing ? new Texturing(*other.texturing) : 0),
	tex_names(other.tex_names),
	back_faces(other.back_faces)
{ }

RenderPass &RenderPass::operator=(const RenderPass &other)
{
	shprog = other.shprog;
	shdata = other.shdata ? new ProgramData(*other.shdata) : 0;
	material = other.material;
	texturing = other.texturing ? new Texturing(*other.texturing) : 0;
	tex_names = other.tex_names;
	back_faces = other.back_faces;
	return *this;
}

RenderPass::~RenderPass()
{
	delete texturing;
}

void RenderPass::set_shader_program(const Program *prog, const ProgramData *data)
{
	shprog = prog;
	shdata = data;
	shdata.keep();
}

void RenderPass::set_material(const Material *mat)
{
	material = mat;
	material.keep();
}

void RenderPass::set_texture(unsigned index, const Texture *tex)
{
	if(!texturing)
		texturing = new Texturing;

	texturing->attach(index, *tex);
}

int RenderPass::get_texture_index(const string &n) const
{
	map<string, unsigned>::const_iterator i = tex_names.find(n);
	if(i==tex_names.end())
		return -1;
	return i->second;
}

void RenderPass::apply(Renderer &renderer) const
{
	renderer.set_texturing(texturing);
	renderer.set_material(material.get());
	renderer.set_shader_program(shprog, shdata.get());
	renderer.set_reverse_winding(back_faces);
}


RenderPass::Loader::Loader(RenderPass &p):
	DataFile::CollectionObjectLoader<RenderPass>(p, 0)
{
	init();
}

RenderPass::Loader::Loader(RenderPass &p, Collection &c):
	DataFile::CollectionObjectLoader<RenderPass>(p, &c)
{
	init();
}

void RenderPass::Loader::init()
{
	add("shader",   &RenderPass::shprog);
	add("material", &Loader::material_inline);
	add("material", &Loader::material);
	add("back_faces",&RenderPass::back_faces);
	add("texunit",  &Loader::texunit);
	add("texunit",  &Loader::texunit_named);
	add("uniforms", &Loader::uniforms);
}

void RenderPass::Loader::material_inline()
{
	RefPtr<Material> mat = new Material;
	if(coll)
		load_sub(*mat, get_collection());
	else
		load_sub(*mat);
	obj.material = mat;
}

void RenderPass::Loader::material(const string &name)
{
	obj.material = &get_collection().get<Material>(name);
	obj.material.keep();
}

void RenderPass::Loader::texunit(unsigned i)
{
	if(!obj.texturing)
		obj.texturing = new Texturing;
	TextureLoader ldr(*obj.texturing, i, coll);
	load_sub_with(ldr);
}

void RenderPass::Loader::texunit_named(unsigned i, const string &n)
{
	texunit(i);
	obj.tex_names[n] = i;
}

void RenderPass::Loader::uniforms()
{
	if(!obj.shprog)
		throw invalid_operation("RenderPass::Loader::uniforms");
	RefPtr<ProgramData> shd = new ProgramData(obj.shprog);
	load_sub(*shd);
	obj.shdata = shd;
}


RenderPass::TextureLoader::TextureLoader(Texturing &t, unsigned i, Collection *c):
	DataFile::CollectionObjectLoader<Texturing>(t, c),
	index(i)
{
	add("texture",   &TextureLoader::texture);
	add("texture2d", &TextureLoader::texture2d);
}

void RenderPass::TextureLoader::finish()
{
	if(tex)
	{
		obj.attach(index, *tex);
		tex.release();
	}
}

void RenderPass::TextureLoader::texture(const string &name)
{
	tex = &get_collection().get<Texture>(name);
	tex.keep();
}

void RenderPass::TextureLoader::texture2d()
{
	tex = new Texture2D;
	if(coll)
		load_sub(static_cast<Texture2D &>(*tex), get_collection());
	else
		load_sub(static_cast<Texture2D &>(*tex));
}

} // namespace GL
} // namespace Msp
